#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>

#include "common.h"
#include "midi.h"
#include "regs.h"
#include "sflist.h"
#include "msgbox.h"
#include "chg_chorus.h"
#include "general_setting.h"

extern GENERAL_SETTING general_setting;

#define gset general_setting

#ifdef _WINDOWS
static gint chg_out_port(GtkWidget *widget, gpointer data);
static gint chg_inp_port(GtkWidget *widget, gpointer data);
#endif

static gint dialog_general_settings_ok(GtkWidget *widget, gpointer data);
static gint dialog_general_settings_cancel(GtkWidget *widget, gpointer data);


     
static GtkWidget *dlg_div_setting;
static GtkWidget *spin_ctlrate;
static GtkWidget *spin_reverb;
static GtkWidget *spin_chorus;
static GtkWidget *spin_voices;
static GtkWidget *entry_timidity_location;
static GtkWidget *entry_timidity_options;

static GtkWidget *entry_eca_location;
static GtkWidget *entry_eca_options;
static GtkWidget *button_eca_enable;


static GtkAdjustment *adj_timi_ctlrate;
static GtkAdjustment *adj_timi_reverb;
static GtkAdjustment *adj_timi_chorus;
static GtkAdjustment *adj_timi_voices;

static int inpport;
static int outport[KB_NUM];



/*
  
*/
gint dialog_general_settings(GtkWidget *widget, gpointer data)
{
  GtkWidget  *button;
  GtkWidget *label;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *frame;
  GtkWidget *notebook;
  GtkWidget *frame_page;
  GtkWidget *vbox_page;
#ifdef _WINDOWS
  GtkWidget *opt_div_ch, *menu_ch;
  int i;
  GtkWidget  *item;
#endif

  general_load_setting();


  // 
  dlg_div_setting = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(dlg_div_setting), "General_Settings");
  gtk_widget_show(dlg_div_setting);


  gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dlg_div_setting)->vbox), 8);


  //////////////////////////////////////////////////////////
  // Notebook

  notebook = gtk_notebook_new();
  gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook), GTK_POS_TOP);
  gtk_widget_show(notebook);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg_div_setting)->vbox), 
		     notebook, TRUE,TRUE,8);


  //========================================================================
  //===============  External Programs Page ================================
  //========================================================================

  frame_page = gtk_frame_new("External Programs");
  gtk_widget_show(frame_page);
  label = gtk_label_new("Ext Progs");
  gtk_notebook_prepend_page(GTK_NOTEBOOK(notebook), frame_page, label);

  vbox_page = gtk_vbox_new(FALSE,8);
  gtk_widget_show(vbox_page);
  gtk_container_add(GTK_CONTAINER(frame_page), vbox_page);


  //////////////////////////////////////////////////////////
  // Timidity ե졼

  frame = gtk_frame_new("TiMidity++");
  gtk_widget_show(frame);
  
  gtk_box_pack_start(GTK_BOX(vbox_page), 
		     frame, TRUE,TRUE,8);

  vbox = gtk_vbox_new(FALSE, 8); //Timidity++ γǼ
  gtk_widget_show(vbox);
  gtk_container_add(GTK_CONTAINER(frame), vbox);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 8);

  ////////////////////////////////////////////////////////////
  //  timidity¹ԥե
  hbox = gtk_hbox_new(FALSE, 4); //٥ȥȥǼ
  gtk_widget_show(hbox);
  
  label = gtk_label_new("Location");
  gtk_widget_show(label);

  entry_timidity_location = gtk_entry_new();
  gtk_widget_show(entry_timidity_location);
  gtk_entry_set_text(GTK_ENTRY(entry_timidity_location),
		     gset.timi_location);

  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE,TRUE,0);
  gtk_box_pack_start(GTK_BOX(hbox), entry_timidity_location, TRUE,TRUE,0);

  gtk_box_pack_start(GTK_BOX(vbox), 
		     hbox, TRUE,TRUE,0);

  ////////////////////////////////////////////////////////////
  //  ȥ쥤
  hbox = gtk_hbox_new(FALSE, 4); //٥ȥȥǼ
  gtk_widget_show(hbox);
  
  label = gtk_label_new("Control rate");
  gtk_widget_show(label);

  adj_timi_ctlrate = (GtkAdjustment *) gtk_adjustment_new (gset.timi_ctlrate,
							   1.0, 127.0, 1.0,
							   5.0, 0.0);

  spin_ctlrate = gtk_spin_button_new (adj_timi_ctlrate, 0, 0);
  gtk_widget_show( spin_ctlrate);

  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE,TRUE,0);
  gtk_box_pack_start(GTK_BOX(hbox), spin_ctlrate, TRUE,TRUE,0);

  gtk_box_pack_start(GTK_BOX(vbox), 
		     hbox, TRUE,TRUE,0);

  ////////////////////////////////////////////////////////////
  //  reverb
  hbox = gtk_hbox_new(FALSE, 4); //٥ȥȥǼ
  gtk_widget_show(hbox);
  
  label = gtk_label_new("Reverb");
  gtk_widget_show(label);

  adj_timi_reverb = (GtkAdjustment *) gtk_adjustment_new (gset.timi_reverb,
							  0.0, 127.0, 1.0,
							  5.0, 0.0);

  spin_reverb = gtk_spin_button_new (adj_timi_reverb, 0, 0);
  gtk_widget_show( spin_reverb );

  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE,TRUE,0);
  gtk_box_pack_start(GTK_BOX(hbox), spin_reverb, TRUE,TRUE,0);

  gtk_box_pack_start(GTK_BOX(vbox), 
		     hbox, TRUE,TRUE,0);

  ////////////////////////////////////////////////////////////
  //  Chorus
  hbox = gtk_hbox_new(FALSE, 4); //٥ȥȥǼ
  gtk_widget_show(hbox);
  
  label = gtk_label_new("Chorus");
  gtk_widget_show(label);

  adj_timi_chorus = (GtkAdjustment *) gtk_adjustment_new (gset.timi_chorus,
							  0.0, 127.0, 1.0,
							  5.0, 0.0);

  spin_chorus = gtk_spin_button_new (adj_timi_chorus, 0, 0);
  gtk_widget_show( spin_chorus );

  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE,TRUE,0);
  gtk_box_pack_start(GTK_BOX(hbox), spin_chorus, TRUE,TRUE,0);

  gtk_box_pack_start(GTK_BOX(vbox), 
		     hbox, TRUE,TRUE,0);


  ////////////////////////////////////////////////////////////
  //  Max voices
  hbox = gtk_hbox_new(FALSE, 4); //٥ȥȥǼ
  gtk_widget_show(hbox);
  
  label = gtk_label_new("Max Voices");
  gtk_widget_show(label);

  adj_timi_voices = (GtkAdjustment *) gtk_adjustment_new (gset.timi_voices,
							  0.0, 511.0, 1.0,
							  5.0, 0.0);

  spin_voices = gtk_spin_button_new (adj_timi_voices, 0, 0);
  gtk_widget_show( spin_voices );

  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE,TRUE,0);
  gtk_box_pack_start(GTK_BOX(hbox), spin_voices, TRUE,TRUE,0);

  gtk_box_pack_start(GTK_BOX(vbox), 
		     hbox, TRUE,TRUE,0);


  ////////////////////////////////////////////////////////////
  //  timidity another options
  hbox = gtk_hbox_new(FALSE, 4); //٥ȥȥǼ
  gtk_widget_show(hbox);
  
  label = gtk_label_new("Another options");
  gtk_widget_show(label);

  entry_timidity_options = gtk_entry_new();
  gtk_widget_show(entry_timidity_options);
  gtk_entry_set_text(GTK_ENTRY(entry_timidity_options),
		     gset.timi_options);

  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE,TRUE,0);
  gtk_box_pack_start(GTK_BOX(hbox), entry_timidity_options, TRUE,TRUE,0);

  gtk_box_pack_start(GTK_BOX(vbox), 
		     hbox, TRUE,TRUE,0);




  //////////////////////////////////////////////////////////
  // EcaSound ե졼

  frame = gtk_frame_new("Ecasound + Freeverb");
  gtk_widget_show(frame);



  gtk_box_pack_start(GTK_BOX(vbox_page), 
		     frame, TRUE,TRUE,0);

  vbox = gtk_vbox_new(FALSE, 8); //ecasound++ γǼ
  gtk_widget_show(vbox);
  gtk_container_add(GTK_CONTAINER(frame), vbox);

  gtk_container_set_border_width(GTK_CONTAINER(vbox), 8);


  ////////////////////////////////////////////////////////////
  //  ecasound ¹ԥե
  hbox = gtk_hbox_new(FALSE, 4); //٥ȥȥǼ
  gtk_widget_show(hbox);
  
  label = gtk_label_new("Location");
  gtk_widget_show(label);

  entry_eca_location = gtk_entry_new();
  gtk_widget_show(entry_eca_location);
  gtk_entry_set_text(GTK_ENTRY(entry_eca_location),
		     gset.eca_location);

  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE,TRUE,0);
  gtk_box_pack_start(GTK_BOX(hbox), entry_eca_location, TRUE,TRUE,0);

  gtk_box_pack_start(GTK_BOX(vbox), 
		     hbox, TRUE,TRUE,0);


  ////////////////////////////////////////////////////////////
  //  timidity another options
  hbox = gtk_hbox_new(FALSE, 4); //٥ȥȥǼ
  gtk_widget_show(hbox);
  
  label = gtk_label_new("Options");
  gtk_widget_show(label);

  entry_eca_options = gtk_entry_new();
  gtk_widget_show(entry_eca_options);
  gtk_entry_set_text(GTK_ENTRY(entry_eca_options),
		     gset.eca_options);

  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE,TRUE,0);
  gtk_box_pack_start(GTK_BOX(hbox), entry_eca_options, TRUE,TRUE,0);

  gtk_box_pack_start(GTK_BOX(vbox), 
		     hbox, TRUE,TRUE,0);

  ////////////////////////////////////////////////////////////
  //  timidity another options
  button_eca_enable = gtk_check_button_new_with_label("Enable ecasound+freeverb");
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_eca_enable), gset.eca_enable);
  gtk_widget_show(button_eca_enable);
  gtk_box_pack_start(GTK_BOX(vbox), 
		     button_eca_enable, TRUE,TRUE,0);



  //========================================================================
  //===============  Events  Page ==========================================
  //========================================================================

  frame_page = gtk_frame_new("MIDI Events");
  gtk_widget_show(frame_page);
  label = gtk_label_new("Events");
  gtk_notebook_prepend_page(GTK_NOTEBOOK(notebook), frame_page, label);

  vbox_page = gtk_vbox_new(FALSE,8);
  gtk_widget_show(vbox_page);
  gtk_container_add(GTK_CONTAINER(frame_page), vbox_page);



#ifdef _WINDOWS
  
  //========================================================================
  //===============  MIDI Ports Page =======================================
  //========================================================================

  frame_page = gtk_frame_new("MIDI Ports");
  gtk_widget_show(frame_page);
  label = gtk_label_new("MIDI Ports");
  gtk_notebook_prepend_page(GTK_NOTEBOOK(notebook), frame_page, label);

  vbox_page = gtk_vbox_new(FALSE,8);
  gtk_widget_show(vbox_page);
  gtk_container_add(GTK_CONTAINER(frame_page), vbox_page);

  //////////////////////////////////////////////////////////
  //  INPUT frame

  frame = gtk_frame_new("INPUT");
  gtk_widget_show(frame);
  
  gtk_box_pack_start(GTK_BOX(vbox_page), 
		     frame, TRUE,TRUE,8);

  vbox = gtk_vbox_new(FALSE, 8); //Input port++ γǼ
  gtk_widget_show(vbox);
  gtk_container_add(GTK_CONTAINER(frame), vbox);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 8);

  //////////////////////////////////////////////////////////
  // Input port selection menu

  hbox = gtk_hbox_new(FALSE, 4); //٥ȥץܥǼ
  gtk_widget_show(hbox);
  
  label = gtk_label_new("Input port");
  gtk_widget_show(label);

  opt_div_ch =gtk_option_menu_new();
  menu_ch = gtk_menu_new();

  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE,TRUE,0);
  gtk_box_pack_start(GTK_BOX(hbox), opt_div_ch, TRUE,TRUE,0);

  // ġ򥢥ƥ
  for (i=0; i<=midiInGetNumDevs(); i++) {
    MIDIINCAPS InCaps;
    if (i==0) {
      item = gtk_menu_item_new_with_label("NoUse");
    } else {
      midiInGetDevCaps(i, (LPMIDIINCAPS) &InCaps,sizeof(InCaps));
      strover80hcut((char*)InCaps.szPname);
      item = gtk_menu_item_new_with_label((char*)InCaps.szPname);
    }
    gtk_menu_append(GTK_MENU(menu_ch), item);
    gtk_signal_connect(GTK_OBJECT(item), "activate",
                       GTK_SIGNAL_FUNC(chg_inp_port), (gpointer)i);
    gtk_widget_show(item);
    
  }
  // ͤ
  inpport = general_get_input_dev();
  gtk_menu_set_active(GTK_MENU(menu_ch), inpport);

  // ˥塼򥪥ץܥȹ
  gtk_option_menu_set_menu(GTK_OPTION_MENU(opt_div_ch), menu_ch);
  gtk_widget_show(opt_div_ch);

  gtk_box_pack_start(GTK_BOX(vbox), 
		     hbox, TRUE,FALSE,0);

  //////////////////////////////////////////////////////////
  //  Loopback frame

  frame = gtk_frame_new("Loopback");
  gtk_widget_show(frame);
  
  gtk_box_pack_start(GTK_BOX(vbox_page), 
		     frame, TRUE,TRUE,8);

  vbox = gtk_vbox_new(FALSE, 8); //Input port++ γǼ
  gtk_widget_show(vbox);
  gtk_container_add(GTK_CONTAINER(frame), vbox);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 8);

  //////////////////////////////////////////////////////////
  // Loopback port selection menu

  for (i=0; i<KB_NUM; i++) {
    int j;
    char labelstr[32];
    
    hbox = gtk_hbox_new(FALSE, 4); //٥ȥץܥǼ
    gtk_widget_show(hbox);
    
    sprintf(labelstr, "Loopback device #%d", i+1);
    label = gtk_label_new(labelstr);
    gtk_widget_show(label);
    
    opt_div_ch =gtk_option_menu_new();
    menu_ch = gtk_menu_new();
    
    gtk_box_pack_start(GTK_BOX(hbox), label, TRUE,TRUE,0);
    gtk_box_pack_start(GTK_BOX(hbox), opt_div_ch, TRUE,TRUE,0);
    
    // ġ򥢥ƥ
    for (j=0; j<=midiOutGetNumDevs(); j++) {
      MIDIOUTCAPS OutCaps;
      if (j==0) {
        item = gtk_menu_item_new_with_label("NoUse");
      } else {
        midiOutGetDevCaps(j-1, (LPMIDIOUTCAPS) &OutCaps,sizeof(OutCaps));
        strover80hcut((char*)OutCaps.szPname);
        item = gtk_menu_item_new_with_label((char*)OutCaps.szPname);
      }
      gtk_menu_append(GTK_MENU(menu_ch), item);
      gtk_signal_connect(GTK_OBJECT(item), "activate",
                         GTK_SIGNAL_FUNC(chg_out_port), (gpointer)(i*100+j));
      gtk_widget_show(item);
      
    }
    // ͤ
    outport[i] = general_get_div_loopback_dev(i);
    gtk_menu_set_active(GTK_MENU(menu_ch), outport[i]);
    
    // ˥塼򥪥ץܥȹ
    gtk_option_menu_set_menu(GTK_OPTION_MENU(opt_div_ch), menu_ch);
    gtk_widget_show(opt_div_ch);
    
    gtk_box_pack_start(GTK_BOX(vbox), 
                       hbox, TRUE,FALSE,0);
  }
  
#endif

  // OKܥ
  button = gtk_button_new_with_label("OK");
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg_div_setting)->action_area), 
		     button, TRUE,FALSE,3);
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
  		     GTK_SIGNAL_FUNC(dialog_general_settings_ok), data);
  gtk_widget_show(button);

  // CANCELܥ
  button = gtk_button_new_with_label("CANCEL");
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg_div_setting)->action_area), 
		     button, TRUE,FALSE,3);
  gtk_signal_connect(GTK_OBJECT(button), "clicked", 
  		     GTK_SIGNAL_FUNC(dialog_general_settings_cancel), 
  		     data);
  gtk_widget_show(button);
 
  gtk_window_set_default_size     (GTK_WINDOW(dlg_div_setting), 200,100);
  gtk_grab_add(GTK_WIDGET(dlg_div_setting));
  return 0;
}


///////////////////////////////////////////////////////////////////////////////

#ifdef _WINDOWS

//  ϥǥХѹ
static gint chg_inp_port(GtkWidget *widget, gpointer data)
{
  inpport = (int)data;
  return 0;
}


// 롼ץХåǥХѹ
// data = div*100 + port
static gint chg_out_port(GtkWidget *widget, gpointer data)
{
  int div = (int)data/100;
  int port  = (int)data % 100;
  outport[div] = port;
  return 0;
}

#endif

//  OK
static gint dialog_general_settings_ok(GtkWidget *widget, gpointer data)
{
  int i;
  strcpy(gset.timi_options, gtk_entry_get_text(GTK_ENTRY(entry_timidity_options)));

  gset.timi_ctlrate = (int)GTK_ADJUSTMENT(adj_timi_ctlrate)->value;
  gset.timi_reverb = (int)GTK_ADJUSTMENT(adj_timi_reverb)->value;
  gset.timi_chorus = (int)GTK_ADJUSTMENT(adj_timi_chorus)->value;
  gset.timi_voices = (int)GTK_ADJUSTMENT(adj_timi_voices)->value;

  general_set_timidity_location(gtk_entry_get_text(GTK_ENTRY(entry_timidity_location)));
  
  strcpy(gset.eca_location, gtk_entry_get_text(GTK_ENTRY(entry_eca_location)));
  strcpy(gset.eca_options, gtk_entry_get_text(GTK_ENTRY(entry_eca_options)));

  gset.eca_enable = GTK_TOGGLE_BUTTON(button_eca_enable)->active;

  for (i=0;i<KB_NUM;i++) {
    general_set_div_loopback_dev(i, outport[i]);
  }
  general_set_input_dev(inpport);

  general_save_setting();

  midi_launch_timidity_and_connect(general_get_soundfont_path(regs_get_soundfont_name(),FALSE));
  midi_set_preset();

  gtk_widget_destroy(dlg_div_setting);


  return 0;
}

//   󥻥
static gint dialog_general_settings_cancel(GtkWidget *widget, gpointer data)
{
  gtk_widget_destroy(dlg_div_setting);
  return 0;
}




