/*****************************************
  Virtual Pipe Organ for Linux (main)
  By Atsushi Ueda <a-ueda@japan-net.ne.jp>
******************************************/



#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>

#ifdef _LINUX
#include <pthread.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/joystick.h>
#endif

#include "common.h"
#include "midi.h"
//#include "ini.h"
#include "regs.h"
#include "sflist.h"
#include "msgbox.h"
#include "chg_chorus.h"
#include "dialog_general_settings.h"
#include "dialog_select_sffile.h"
#include "dialog_adj_pitch.h"
#include "general_setting.h"
#include "main.h"


// ᥤ󥦥ɥ
GtkWidget *main_window =NULL;

// ȥåץꥹ
GtkWidget *list[KB_NUM];

// ץܥ [div][to-div]
GtkWidget *button_cp[KB_NUM][KB_NUM];

// ȥߥ
GtkWidget *button_tr[KB_NUM];

// ̾
GtkWidget *label_div_name[KB_NUM];

GtkWidget *memory_button;
GtkWidget *chorus_button;
int cancel_set_combination = 0;

#ifdef _LINUX
pthread_t thread_midi;
pthread_t thread_js;
#endif


typedef struct {
  int keyb;
  int stop;
} RIGHTSEL;


int set_stop_name_list()
{
  char *dummy[]={""};
  char *str[2];
  int i,j;
  for (i=0; i<KB_NUM; i++) {
    // ̾
    char label[64];
    sprintf(label, "%d: %s", i+1, regs_get_werk_name(i));
    gtk_label_set_text(GTK_LABEL(label_div_name[i]), label);
    gtk_clist_clear (GTK_CLIST(list[i]));
    gtk_clist_set_selection_mode(GTK_CLIST(list[i]),GTK_SELECTION_MULTIPLE);

    for (j=0; j<STOP_NUM; j++) {
      gtk_clist_append(GTK_CLIST(list[i]), dummy);
    }

    for (j=0; j<STOP_NUM; j++) {
      str[0] = regs_get_stop_name(i,j);
      //gtk_clist_append(GTK_CLIST(list[i]), dummy);
      gtk_clist_set_text(GTK_CLIST(list[i]), j, 0, regs_get_stop_name(i,j));
      gtk_clist_set_selectable   (GTK_CLIST(list[i]), j, strlen(str[0])?1:0);
    }
    gtk_clist_set_selection_mode(GTK_CLIST(list[i]),GTK_SELECTION_MULTIPLE);
  }
  return 0;
}


/*
    Υȥå֤̾CListȿ
*/
void display_combination()
{
  int i,j;

  //֤ȿǤȥޥ򤷤Ȥư򤷤ƤޤΤǡ
  //򥭥󥻥뤹롣
  cancel_set_combination = 1; 

  for (i=0; i<KB_NUM; i++) {
    // stops
    for (j=0; j<STOP_NUM; j++) {
      //printf("    === %d %d %d\n", i,j,regs_get_current_stop_select(i,j));
      if (regs_get_current_stop_select(i,j)) {
	gtk_clist_select_row( GTK_CLIST(list[i]), j, 0);
	//printf("---%d,%d\n",i,j);
      } else {
	gtk_clist_unselect_row( GTK_CLIST(list[i]), j, 0);
      }
    }
    // coupler
    for (j=0; j<KB_NUM; j++) {
      if (i!=j) {
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_cp[i][j]),
				     regs_get_current_coupler_state(i,j));
      }
    }
    // tremulant
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_tr[i]),
				 regs_get_current_tremulant_state(i));
    chg_tremulant(button_tr[i], (gpointer)i);
  }
  cancel_set_combination = 0;
}




/*
  쥸ȥ졼ɤ߹
  tbl_fn : 쥸ȥ졼ơ֥ե̾
  fl_screen_chg: FALSEȲ̤ι䡢SF̵䤤碌Ԥʤ
*/ 
static int load_registration_table(char *tbl_fn, int flg_screen_chg)
{
  int rc = TRUE;
  int flgSffileSelected = FALSE;
  
  char soundfont_name_tmp[1024];
  char soundfont_name[1024];
  char soundfont_fn[1024];
  char *soundfont_fn_tmp;

  GetPrivateProfileString("General", "SoundFont", 
			  "__NG__", soundfont_name_tmp,1024, tbl_fn);
  if (strcmp(soundfont_name_tmp,"__NG__")!=0) {
    csv_read(soundfont_name_tmp,soundfont_name,0); //remove double quotation
    soundfont_fn_tmp = general_get_soundfont_path(soundfont_name, flg_screen_chg);
    strcpy(soundfont_fn, soundfont_fn_tmp?soundfont_fn_tmp:"");
  } else {
    rc= FALSE;
  }

  puts("(1)");

  if (rc) {
    for (;;) {
      if (sflist_load(soundfont_fn)==FALSE) {
	char msg[1000];
	if (flg_screen_chg == FALSE) {
	  break;
	}
	sprintf(msg,"SoundFont file %s does not exists. Please select.",
		soundfont_fn);
	msgbox(msg, "File not found.", "OK", 0);
	if (file_dialog(soundfont_fn, 
			"Select SoundFont file", soundfont_fn)==TRUE) {
	  flgSffileSelected = TRUE;
	} else {
	  rc=FALSE;
	  break;
	}
      } else {
        char cfg_fn[1000];
        sprintf(cfg_fn, "%s/tmp.cfg", general_get_rc_dir());
        sflist_make_timidity_config_file(soundfont_fn, cfg_fn);
        break;
      }
    }
  }

  if (rc) {
    char title[1000];
    regs_read_registration_table_file(tbl_fn);

    // vpoե˵ҤƤɥե̵̾ʾ⤢Τǡ
    // դʤä򤷤Τͥ褹롣
    regs_set_soundfont_name(sflist_getSFName());
    general_set_soundfont_name(sflist_getSFName(), sflist_getSFFileName());

    if (main_window) {
      sprintf(title, "Linux Pipe Organ : %s", tbl_fn);
      gtk_window_set_title(GTK_WINDOW(main_window),title);
    }

  }

  return rc;

  if (flgSffileSelected == TRUE) {
    //regs_set_sffile_name(soundfont_fn);
  }

  if (flg_screen_chg == TRUE) {
    set_stop_name_list();
    display_combination();
    midi_set_preset();
  }
  return rc;
}







/*
    ̾Υȥå
    ȿǤ
*/
void set_combination()
{
  GList *selection;
  gint row;
  int i,j;
  if (cancel_set_combination) {
    return;
  }
  for (i=0; i<KB_NUM; i++) {
    for (j=0; j<STOP_NUM; j++) {
      regs_set_current_stop_select(i,j,0);
    }
    selection = GTK_CLIST(list[i])->selection;
    while(selection) {
      row = GPOINTER_TO_INT(selection->data);
      // printf("   %d,%d\n",i,row);
      regs_set_current_stop_select(i,row,1);
      selection = g_list_next(selection);
    }
  }
}



///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

static GtkWidget* fsel_reg;


/*
  쥸ȥ졼ɤ߹ߥե륻쥯ȥ
*/
gint open_reg(GtkWidget *widget, gpointer data)
{
  char fn[1000];
  char *sf_path;

  if (file_dialog(fn, "Open Registration File", general_get_recent_registration_file())==TRUE) {
    if (load_registration_table(fn, TRUE) == FALSE) {
      return 0;
    }

    general_set_recent_registration_file(fn);
    general_save_setting();

    sf_path = general_get_soundfont_path(regs_get_soundfont_name(), TRUE);

    printf("Selected: %s\n", sf_path);

    if (midi_launch_timidity_and_connect(sf_path) < 0) {
      fprintf(stderr, "Couldn't connect to timidity.\n");
      exit(1);
    }
    set_stop_name_list();
    display_combination();
    midi_set_preset();
  }
  
  
  return 0;
}

/*
  쥸ȥ졼¸OK
*/
void file_ok_save_reg (GtkWidget *w, GtkFileSelection *fs)
{
  char fn[1000];
  const char *fn0 = gtk_file_selection_get_filename (GTK_FILE_SELECTION (fs));

  strcpy(fn, fn0);
  if (!strchr(fn,'.')) {
    strcat(fn,".vpo");
  }

  general_set_recent_registration_file(fn);
  general_save_setting();

  if (access(fn, R_OK)==0) {
    char str[1000];
    sprintf(str, "File <%s> already exists. Proceed?", fn);
    if (msgbox(str, "VPO","YES|NO", 1)==1) {
      gtk_widget_destroy(fsel_reg);
      return;
    }
  }
  regs_write_registration_table_file(fn);
  gtk_widget_destroy(fsel_reg);

}


/*
  쥸ȥ졼¸ե륻쥯ȥФ
*/
gint save_reg(GtkWidget *widget, gpointer data)
{
  fsel_reg = gtk_file_selection_new ("Save Registration File");
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fsel_reg)->ok_button),
		      "clicked", (GtkSignalFunc) file_ok_save_reg, fsel_reg );
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(fsel_reg)->cancel_button),
			     "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy),
			     GTK_OBJECT (fsel_reg));
  gtk_file_selection_set_filename (GTK_FILE_SELECTION (fsel_reg), 
				   "");
  gtk_widget_show (fsel_reg);
  gtk_grab_add(GTK_WIDGET(fsel_reg));

  return 0;
}



///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

static GtkWidget *dlg_chg_velocity;
static GtkWidget *scale_velocity;
static GtkObject *adj_velocity;





// velocity ѹ OK
gint chg_velocity_dialog_ok(GtkWidget *widget, gpointer data)
{
  RIGHTSEL *rsel;
  rsel =(RIGHTSEL*)data;
  regs_set_current_stop_velocity(rsel->keyb, 
				 rsel->stop,
				 (int)GTK_ADJUSTMENT(adj_velocity)->value);
  gtk_widget_destroy(dlg_chg_velocity);

  return 0;
}

// velocityѹ 󥻥
gint chg_velocity_dialog_cancel(GtkWidget *widget, gpointer data)
{
  gtk_widget_destroy(dlg_chg_velocity);

  return 0;
}


/*
  ȥåñ̤Υƥ
*/
gint chg_velocity_dialog(GtkWidget *widget, gpointer data)
{
  GtkWidget *button;
  RIGHTSEL *rsel;

  rsel = (RIGHTSEL*)data;

  dlg_chg_velocity = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(dlg_chg_velocity), "Change Velocity");
 
  gtk_widget_show(dlg_chg_velocity);

  adj_velocity = gtk_adjustment_new (regs_get_current_stop_velocity(rsel->keyb,rsel->stop),
				     0.0, 128.0, 1.0, 1.0, 1.0);
  scale_velocity = gtk_hscale_new (GTK_ADJUSTMENT (adj_velocity));
  gtk_widget_set_usize (GTK_WIDGET (scale_velocity), 150, 30);
  gtk_range_set_update_policy (GTK_RANGE (scale_velocity), 
			       GTK_UPDATE_DELAYED);
  gtk_scale_set_digits (GTK_SCALE (scale_velocity), 0); //ʲ
  gtk_scale_set_draw_value (GTK_SCALE (scale_velocity), TRUE); //ɽ
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg_chg_velocity)->vbox), 
		     scale_velocity, TRUE,TRUE,0);
  gtk_widget_show(scale_velocity);


  button = gtk_button_new_with_label("OK");
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg_chg_velocity)->action_area), 
		     button, TRUE,FALSE,3);
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
		     GTK_SIGNAL_FUNC(chg_velocity_dialog_ok), (gpointer)rsel);
  gtk_widget_show(button);

  button = gtk_button_new_with_label("CANCEL");
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg_chg_velocity)->action_area), 
		     button, TRUE,FALSE,3);
  gtk_signal_connect(GTK_OBJECT(button), "clicked", 
		     GTK_SIGNAL_FUNC(chg_velocity_dialog_cancel), 
		     (gpointer)rsel);
  gtk_widget_show(button);
 
  gtk_window_set_default_size     (GTK_WINDOW(dlg_chg_velocity), 200,100);
  gtk_grab_add(GTK_WIDGET(dlg_chg_velocity));
  return 0;
}





static GtkWidget *dlg_chg_preset, *clist_chg_preset;


// ѹ OK
gint chg_preset_dialog_ok(GtkWidget *widget, gpointer data)
{
  RIGHTSEL *rsel;
  GList *selection;

  rsel =(RIGHTSEL*)data;
  
  selection = GTK_CLIST(clist_chg_preset)->selection;
  if (selection) {
    int bank, preset, pre_bank, pre_preset, sel;
    sel = GPOINTER_TO_INT(selection->data);
    regs_get_preset(rsel->keyb, rsel->stop, &pre_bank, &pre_preset);
    sflist_getname_n(sel, &bank, &preset);
    if (bank!=pre_bank || preset!=pre_preset) {
      regs_set_preset(rsel->keyb, rsel->stop, bank, preset);
      set_stop_name_list();
      display_combination();
      midi_set_preset();
    }
  }
  gtk_widget_destroy(dlg_chg_preset);

  return 0;
}

// ѹ 󥻥
gint chg_preset_dialog_cancel(GtkWidget *widget, gpointer data)
{
  gtk_widget_destroy(dlg_chg_preset);

  return 0;
}


/*
  ѹ
*/
gint chg_preset_dialog(GtkWidget *widget, gpointer data)
{
  int cur_bank,cur_preset; // ѹоݤΡߤβ
  GtkWidget  *swin, *button;
  int i;
  RIGHTSEL *rsel;
  int selection = -1;

  rsel = (RIGHTSEL*)data;
  regs_get_preset(rsel->keyb, rsel->stop, &cur_bank, &cur_preset);
  dlg_chg_preset = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(dlg_chg_preset), "Change Preset");
 
  clist_chg_preset = gtk_clist_new(1);

  swin = gtk_scrolled_window_new(NULL,NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(swin), 
                                 GTK_POLICY_NEVER, GTK_POLICY_ALWAYS); 
  gtk_widget_show(swin);

  // ꥹȥܥå˲̾
  for (i=0; i<sflist_getnum(); i++) {
    char *dummy[]={""};
    char *str[2];
    char *name;
    int bank, preset;
    char tmpname[255] ,*p;
    // ̾
    name = sflist_getname_n(i, &bank, &preset);
    // Х󥯡ץꥻåȡ̾ʸ
    sprintf(tmpname,"%d,%3d     %s", bank, preset, name);
    // ʸ椫8ӥåʸ
    for (p=tmpname; *p; p++) {
      if ( (unsigned char)*p >= 0x80) *p=' ';
    }
    // ̾Ĥƥ
    str[0] = tmpname;
    // ꥹȥܥå̾Ĥƥɲ
    gtk_clist_append(GTK_CLIST(clist_chg_preset), dummy);
    gtk_clist_set_text(GTK_CLIST(clist_chg_preset), i, 0, tmpname);
    // ߤ֤ˤ
    if (bank==cur_bank && preset==cur_preset) {
      selection = i;
    }
  }
  gtk_container_add(GTK_CONTAINER(swin), clist_chg_preset);
    

  gtk_widget_show(clist_chg_preset);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg_chg_preset)->vbox), swin, TRUE,TRUE,0);
  gtk_widget_show(dlg_chg_preset);

  // ꤬򤵤줿֤ˤ
  if (selection >= 0) {
    gtk_clist_select_row (GTK_CLIST(clist_chg_preset), selection,0);
    gtk_clist_moveto(GTK_CLIST(clist_chg_preset), selection, 0,0,0);
  }


  button = gtk_button_new_with_label("OK");
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg_chg_preset)->action_area), button, TRUE,FALSE,3);
  gtk_signal_connect(GTK_OBJECT(button), "clicked",GTK_SIGNAL_FUNC(chg_preset_dialog_ok), (gpointer)rsel);
  gtk_widget_show(button);

  button = gtk_button_new_with_label("CANCEL");
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg_chg_preset)->action_area), button, TRUE,FALSE,3);
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(chg_preset_dialog_cancel), (gpointer)rsel);
  gtk_widget_show(button);
 
  gtk_window_set_default_size     (GTK_WINDOW(dlg_chg_preset), 250,300);
  gtk_grab_add(GTK_WIDGET(dlg_chg_preset));
  return 0;
}

/*
  
*/
gint insert_stop(GtkWidget *widget, gpointer data)
{
  RIGHTSEL rsel = *(RIGHTSEL*)data;
  regs_insert(rsel.keyb, rsel.stop);
  set_stop_name_list();
  display_combination();
  midi_set_preset();
  return 0;
}


/*
  
*/
gint delete_stop(GtkWidget *widget, gpointer data)
{
  RIGHTSEL rsel = *(RIGHTSEL*)data;
  printf("delete %d %d\n",rsel.keyb, rsel.stop);
  regs_delete(rsel.keyb, rsel.stop);
  set_stop_name_list();
  display_combination();
  midi_set_preset();
  return 0;
}




/*
  ȥåפ򱦥å
*/
gint clist_right(GtkWidget *widget,  GdkEventButton * event, gpointer data)
{
  static RIGHTSEL rsel;
  GtkWidget *menu, *menu_item;
  int row,col;
  if (!event) return 0;
  if (event->button != 3) return 0; //right click only
  gtk_clist_get_selection_info (GTK_CLIST(list[(int)data]), (int)event->x, (int)event->y, &row, &col);
  rsel.keyb = (int)data;
  rsel.stop = row;
  menu = gtk_menu_new();
  menu_item = gtk_menu_item_new_with_label("Insert");
  gtk_menu_append(GTK_MENU(menu), menu_item);
  gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(insert_stop), &rsel);
  gtk_widget_show(menu_item);
  menu_item = gtk_menu_item_new_with_label("Delete");
  gtk_menu_append(GTK_MENU(menu), menu_item);
  gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(delete_stop), &rsel);
  gtk_widget_show(menu_item);
  menu_item = gtk_menu_item_new_with_label("Change preset");
  gtk_menu_append(GTK_MENU(menu), menu_item);
  gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(chg_preset_dialog), &rsel);
  gtk_widget_show(menu_item);
  menu_item = gtk_menu_item_new_with_label("velocity");
  gtk_menu_append(GTK_MENU(menu), menu_item);
  gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(chg_velocity_dialog), &rsel);
  gtk_widget_show(menu_item);
  gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, 
		  event->button, event->time); 
  return 0;
}


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////



/*
    ӥ͡ܥ󲡲
*/
gint combination_btn(GtkWidget *widget, gpointer data)
{
  if (GTK_TOGGLE_BUTTON (memory_button)->active) {
    gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(memory_button),0);
    regs_memory_registration((int)data);
  } else {
    printf("call %d\n",(int)data);
    regs_call_registration((int)data);
    display_combination();
    midi_set_all_voice_sts();
//    midi_set_stop_volume();
  }
  return 0;
}



/*
    ȥåפ줿ȿ
*/
gint chg_stop(GtkWidget *widget, gpointer data)
{
  set_combination();

  if (cancel_set_combination) return 0; 
  
  midi_set_all_voice_sts();


  midi_set_stop_volume();

  return 0;
}



/*
    ȥåפˤ
 */
gint unselect_all_stop(GtkWidget *widget, gpointer data)
{
  int i,j;
  for (i=0; i<KB_NUM; i++) {
    for (j=0; j<STOP_NUM; j++) {
      regs_set_current_stop_select(i,j,0);
    }
    for (j=0; j<KB_NUM; j++) {
      regs_set_current_coupler_state(i,j, 0);
    }
    regs_set_current_tremulant_state(i,0);
  }
  display_combination();
  midi_set_all_voice_sts();
  midi_set_stop_volume();

  return 0;
}


// 饹ܥ
gint chg_chorus(GtkWidget *widget, gpointer data)
{
  midi_enable_chorus(GTK_TOGGLE_BUTTON(widget)->active);
  printf(" chorus %d\n",GTK_TOGGLE_BUTTON(widget)->active);

  return 0;
}

gint chg_tremulant(GtkWidget *widget, gpointer data)
{
  midi_enable_tremulant((int)data, GTK_TOGGLE_BUTTON(widget)->active);
  regs_set_current_tremulant_state((int)data,GTK_TOGGLE_BUTTON(widget)->active?1:0);
  return 0;
}

// ץ顼ܥ
gint chg_coupler(GtkWidget *widget, gpointer data)
{
  regs_set_current_coupler_state((int)data/10, (int)(data)%10, 
			 GTK_TOGGLE_BUTTON(widget)->active);
  midi_set_all_voice_sts();
  return 0;
}



///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

/*
  
*/



GtkWidget *dlg_div_setting;
GtkObject *adj_div_volume;
GtkObject *adj_transpose;
GtkWidget *scale_div_volume;
GtkWidget *opt_div_ch, *menu_ch;
GtkWidget *entry_div_name;
GtkWidget *spin_transpose;
int div_ch;

//  OK
gint div_setting_dialog_ok(GtkWidget *widget, gpointer data)
{
  general_set_div_channel((int)data, (div_ch==0) ? CH_OFF : div_ch-1);
  regs_set_werk_name((int)data,gtk_entry_get_text(GTK_ENTRY(entry_div_name)));
  set_stop_name_list();
  display_combination();
  general_set_div_transpose((int)data, (int)GTK_ADJUSTMENT(adj_transpose)->value);
  gtk_widget_destroy(dlg_div_setting);
  general_save_setting();
   return 0;
}

//   󥻥
gint div_setting_dialog_cancel(GtkWidget *widget, gpointer data)
{
  gtk_widget_destroy(dlg_div_setting);
  return 0;
}

// ϥͥѹ
gint chg_div_ch(GtkWidget *widget, gpointer data)
{
  div_ch = (int)data;
  return 0;
}

/*
  
*/
gint div_setting(GtkWidget *widget, gpointer data)
{
  GtkWidget  *item, *button;
  GtkWidget *label;
  GtkWidget *hbox;
  int i;
  char title[32];

  // 
  sprintf(title,"KB Setting %d",(int)data + 1);
  dlg_div_setting = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(dlg_div_setting), title);
  gtk_widget_show(dlg_div_setting);

  //////////////////////////////////////////////////////////
  // ϥͥ򥪥ץܥ

  hbox = gtk_hbox_new(FALSE, 4); //٥ȥץܥǼ
  gtk_widget_show(hbox);
  
  label = gtk_label_new("MIDI input channel");
  gtk_widget_show(label);

  opt_div_ch =gtk_option_menu_new();
  menu_ch = gtk_menu_new();

  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE,TRUE,0);
  gtk_box_pack_start(GTK_BOX(hbox), opt_div_ch, TRUE,TRUE,0);

  // ġ򥢥ƥ
  for (i=0; i<=16; i++) {
    char str[32];
    sprintf(str, "CH %d", i);
    item = gtk_menu_item_new_with_label(i?str:"OFF");
    gtk_menu_append(GTK_MENU(menu_ch), item);
    gtk_signal_connect(GTK_OBJECT(item), "activate",
		       GTK_SIGNAL_FUNC(chg_div_ch), (gpointer)i);
    gtk_widget_show(item);
    
  }
  // ͤ
  div_ch = general_get_div_channel((int)data);
  div_ch = (div_ch>=0 && div_ch<=15) ? div_ch+1 : 0;
  gtk_menu_set_active(GTK_MENU(menu_ch), div_ch);

  // ˥塼򥪥ץܥȹ
  gtk_option_menu_set_menu(GTK_OPTION_MENU(opt_div_ch), menu_ch);
  gtk_widget_show(opt_div_ch);

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg_div_setting)->vbox), 
		     hbox, TRUE,TRUE,0);


  ////////////////////////////////////////////////////////////
  //  ̾
  hbox = gtk_hbox_new(FALSE, 4); //٥ȥȥǼ
  gtk_widget_show(hbox);
  
  label = gtk_label_new("Division Name");
  gtk_widget_show(label);

  entry_div_name = gtk_entry_new();
  gtk_widget_show(entry_div_name);
  gtk_entry_set_text(GTK_ENTRY(entry_div_name),
		     regs_get_werk_name((int)data));

  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE,TRUE,0);
  gtk_box_pack_start(GTK_BOX(hbox), entry_div_name, TRUE,TRUE,0);

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg_div_setting)->vbox), 
		     hbox, TRUE,TRUE,0);


  ////////////////////////////////////////////////////////////
  // ܥ塼ѥ饤
  adj_div_volume = gtk_adjustment_new (64,
				      0.0, 128.0, 1.0, 1.0, 1.0);
  scale_div_volume = gtk_hscale_new (GTK_ADJUSTMENT (adj_div_volume));

  gtk_widget_set_usize (GTK_WIDGET (scale_div_volume), 150, 30);
  gtk_range_set_update_policy (GTK_RANGE (scale_div_volume), 
			       GTK_UPDATE_DELAYED);
  gtk_scale_set_digits (GTK_SCALE (scale_div_volume), 0); //ʲ
  gtk_scale_set_draw_value (GTK_SCALE (scale_div_volume), TRUE); //ɽ
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg_div_setting)->vbox), 
		     scale_div_volume, TRUE,TRUE,0);
  //gtk_widget_show(scale_div_volume);


  ////////////////////////////////////////////////////////////
  //  Transpose
  hbox = gtk_hbox_new(FALSE, 4); //٥ȥȥǼ
  gtk_widget_show(hbox);
  
  label = gtk_label_new("Transpose");
  gtk_widget_show(label);

  adj_transpose = gtk_adjustment_new ((double)general_get_div_transpose((int)data),
				      -36.0, 36.0, 1.0,
				      5, 0);
  
  spin_transpose = gtk_spin_button_new ((GtkAdjustment*)adj_transpose, 0, 0);
  gtk_widget_show( spin_transpose );

  gtk_box_pack_start(GTK_BOX(hbox), label, TRUE,TRUE,0);
  gtk_box_pack_start(GTK_BOX(hbox), spin_transpose, TRUE,TRUE,0);

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg_div_setting)->vbox),
                     hbox, TRUE,TRUE,0);

  
  // OKܥ
  button = gtk_button_new_with_label("OK");
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg_div_setting)->action_area), 
		     button, TRUE,FALSE,3);
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
		     GTK_SIGNAL_FUNC(div_setting_dialog_ok), data);
  gtk_widget_show(button);

  // CANCELܥ
  button = gtk_button_new_with_label("CANCEL");
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg_div_setting)->action_area), 
		     button, TRUE,FALSE,3);
  gtk_signal_connect(GTK_OBJECT(button), "clicked", 
		     GTK_SIGNAL_FUNC(div_setting_dialog_cancel), 
		     data);
  gtk_widget_show(button);
 
  gtk_window_set_default_size     (GTK_WINDOW(dlg_div_setting), 200,100);
  gtk_grab_add(GTK_WIDGET(dlg_div_setting));
  return 0;
}





///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


// ΥȥåץꥹȺ
static GtkWidget* MkStopPanel(int div)
{
  GtkWidget *hbox;
  GtkWidget *vbox;
  GtkWidget *swin;
  GtkWidget *button;
  int i;
  char label[64];
  
  // ΤǼľѥå󥰥ܥå
  vbox = gtk_vbox_new(FALSE, 4); //4ɥåȤ
  gtk_widget_show(vbox);


  // ̾
  sprintf(label, "%d: %s", div+1, regs_get_werk_name(div));
  label_div_name[div] = gtk_label_new(label);
  gtk_widget_show(label_div_name[div]);
  gtk_box_pack_start(GTK_BOX(vbox), label_div_name[div],FALSE,FALSE,0);


  swin = gtk_scrolled_window_new(NULL,NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(swin), 
				 GTK_POLICY_NEVER, GTK_POLICY_NEVER);
  gtk_box_pack_start(GTK_BOX(vbox), swin, FALSE,FALSE,0);
  gtk_widget_show(swin);


  // ꥹȺ
  list[div] = gtk_clist_new(1);
  gtk_container_add(GTK_CONTAINER(swin), list[div]);
  gtk_signal_connect(GTK_OBJECT(list[div]),"select_row",
		     GTK_SIGNAL_FUNC(chg_stop),(gpointer)0);
  gtk_signal_connect(GTK_OBJECT(list[div]),"unselect_row",
		     GTK_SIGNAL_FUNC(chg_stop),(gpointer)0);
  gtk_signal_connect(GTK_OBJECT(list[div]),"button_press_event",
		     GTK_SIGNAL_FUNC(clist_right),(gpointer)div);
  gtk_widget_show(list[div]);

  // ץܥȥꡦIFFܥǼʿѥå󥰥ܥå
  hbox = gtk_hbox_new(FALSE, 4); //4ɥåȤ
  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE,FALSE,0);

  // ץܥ
  for (i=0; i<KB_NUM; i++) {
    if (i!=div) {
      char label[8];
      sprintf(label,"to%d", i+1);
      button_cp[div][i] = gtk_toggle_button_new_with_label(label);
      gtk_signal_connect(GTK_OBJECT(button_cp[div][i]),"toggled",
			 GTK_SIGNAL_FUNC(chg_coupler),(gpointer)(div*10+i));
      gtk_widget_show(button_cp[div][i]);
      gtk_box_pack_start(GTK_BOX(hbox), button_cp[div][i], TRUE,TRUE,0);
    }
  }

  // ȥߥȤܥǼʿѥå󥰥ܥå
  hbox = gtk_hbox_new(FALSE, 4); //4ɥåȤ
  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE,FALSE,0);

  // ȥߥȥܥ
  button_tr[div] = gtk_toggle_button_new_with_label("Tremulant");
  gtk_widget_show(button_tr[div]);
  gtk_signal_connect(GTK_OBJECT(button_tr[div]),"toggled",
		     GTK_SIGNAL_FUNC(chg_tremulant),(gpointer)div);
  gtk_box_pack_start(GTK_BOX(hbox), button_tr[div], TRUE,TRUE,0);

  // ܥ
  button = gtk_button_new_with_label("Setting");
  gtk_widget_show(button);
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(div_setting),(gpointer)div);
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE,TRUE,0);

  return vbox;
}


// ӥ͡ܥ
static GtkWidget* MkCombBtn(int no)
{
  GtkWidget *button;
  char label[8];
  sprintf(label, "%d", no);
  button = gtk_button_new_with_label(label);
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(combination_btn),(gpointer)no);
  gtk_widget_show(button);
  return button;
}

#define JOY_DEV "/dev/js0"

/*
  borrowed from http://lgdc.sunsite.dk/articles/19.html
*/
void joystick_main(void *ptr)
{
#if 0 
  int joy_fd, *axis=NULL, num_of_axis=0, num_of_buttons=0;
  char *button=NULL, name_of_joystick[80];
  struct js_event js;
  
  if( ( joy_fd = open( JOY_DEV , O_RDONLY)) == -1 ) {
    fprintf(stderr, "Couldn't open joystick\n" );
    return ;
  }
  
  ioctl( joy_fd, JSIOCGBUTTONS,&num_of_buttons );
  ioctl( joy_fd, JSIOCGNAME(80),&name_of_joystick );
  
  axis = (int *) calloc( num_of_axis,sizeof( int ) );
  button = (char *) calloc( num_of_buttons,sizeof( char ) );
  
  fcntl( joy_fd, F_SETFL, O_NONBLOCK );  /* use non-blocking mode */
  
  while( 1 ) { /* infinite loop */

    /* read the joystick state */
    read(joy_fd, &js, sizeof(struct js_event));
      
    /* see what to do with the event */
    switch (js.type & ~JS_EVENT_INIT) {
    case JS_EVENT_BUTTON:
      if (button[js.number]==0 && js.value) {
	gdk_threads_enter();
	combination_btn(NULL, (gpointer)(js.number+1));
	gtk_widget_queue_draw (main_window); 
	gdk_threads_leave();
      }
      button [ js.number ] = js.value;
      break;
    }
    
    usleep(50000);
  }
  
  close( joy_fd );
#endif
}


/*
  λ
*/
gint delete_event (GtkWidget *widget, GdkEvent *event, GtkWidget *parent)
{
  char *vpo_default_fn;
  // ߤξ򥳥ӥ͡󣰤¸
  regs_memory_registration(0);
  
  // ǥեȤ registration file ¸
  if ((vpo_default_fn = general_get_default_vpofile())) {  
    regs_write_registration_table_file(vpo_default_fn);
  }
  general_save_setting();
  return FALSE; //windowλ
}




/*
  ᥤ(ɥκᥤ롼)
*/
int main(int argc, char *argv[])
{
  GtkWidget *button;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *menu_bar, *menu, *item;

  int i;
  int client_id=0;

  // GTK
#ifdef _LINUX
  g_thread_init(NULL);
#endif
  gtk_set_locale();
  gtk_init(&argc,&argv);  
   
  general_load_setting();
  regs_init();

  // ޥɰMIDI³ݡȤ
  for (i=1; i<argc; i++) {
    if (strncmp(argv[i],"-p",2)==0) {
      sscanf(argv[i],"-p%d", &client_id);
    }
  }

  if (general_get_default_vpofile()) {
    printf("------1------%s\n", general_get_default_vpofile());
    load_registration_table(general_get_default_vpofile(), FALSE);
  }

  midi_init();
  if (client_id) {
//    midi_init(65);
  }

  midi_open_input();

  if (client_id==0) {
    printf("<%s>\n",regs_get_soundfont_name());
    printf("<%s>\n",general_get_soundfont_path(regs_get_soundfont_name(),FALSE));

    if (midi_launch_timidity_and_connect(general_get_soundfont_path(regs_get_soundfont_name(),FALSE)) ==FALSE) {
      msgbox("Couldn't connect to timidity.","error","OK",0);
      //exit(1);
    }
  }


  // ȥåץ٥륦ɥ
  main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(main_window),"Linux Pipe Organ");

  gtk_signal_connect (GTK_OBJECT (main_window), "delete_event",
		      GTK_SIGNAL_FUNC (delete_event), main_window);
  gtk_signal_connect (GTK_OBJECT (main_window), "destroy",
		      GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
  




  // ľܥå (˥塼Сȥơ֥뤬)
  vbox = gtk_vbox_new(FALSE, 4); //4ɥåȤ
  gtk_container_add(GTK_CONTAINER(main_window), vbox);
  gtk_widget_show(vbox);

  // ˥塼
  menu_bar = gtk_menu_bar_new();
  gtk_box_pack_start(GTK_BOX(vbox), menu_bar, FALSE,FALSE,0);
  gtk_widget_show(menu_bar);
  menu = gtk_menu_new();

  item = gtk_menu_item_new_with_label("Open registration file");
  gtk_menu_append(GTK_MENU(menu), item);
  gtk_signal_connect(GTK_OBJECT(item), "activate",
		     GTK_SIGNAL_FUNC(open_reg), 0);
  gtk_widget_show(item);

  item = gtk_menu_item_new_with_label("Save registration file");
  gtk_menu_append(GTK_MENU(menu), item); 
  gtk_signal_connect(GTK_OBJECT(item), "activate",
		     GTK_SIGNAL_FUNC(save_reg), 0);
  gtk_widget_show(item);

  item = gtk_menu_item_new_with_label("General Settings");
  gtk_menu_append(GTK_MENU(menu), item);
  gtk_signal_connect(GTK_OBJECT(item), "activate",
		     GTK_SIGNAL_FUNC(dialog_general_settings), 0);
  gtk_widget_show(item);

  item = gtk_menu_item_new_with_label("Change Sound Font");
  gtk_menu_append(GTK_MENU(menu), item);
  gtk_signal_connect(GTK_OBJECT(item), "activate",
		     GTK_SIGNAL_FUNC(dialog_select_sffile), 0);
  gtk_widget_show(item);




  item = gtk_menu_item_new_with_label("File");
  gtk_widget_show(item);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);
  gtk_menu_bar_append(GTK_MENU_BAR(menu_bar), item);

  menu = gtk_menu_new();

  item = gtk_menu_item_new_with_label("Chorus level");
  gtk_menu_append(GTK_MENU(menu), item);
  gtk_signal_connect(GTK_OBJECT(item), "activate",
		     GTK_SIGNAL_FUNC(chg_chorus_level_dialog), 0);
  gtk_widget_show(item);

  item = gtk_menu_item_new_with_label("Pitch");
  gtk_menu_append(GTK_MENU(menu), item);
  gtk_signal_connect(GTK_OBJECT(item), "activate",
		     GTK_SIGNAL_FUNC(adj_pitch_dialog), 0);
  gtk_widget_show(item);


  item = gtk_menu_item_new_with_label("Control");
  gtk_widget_show(item);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);
  gtk_menu_bar_append(GTK_MENU_BAR(menu_bar), item);




  //////////////////////////////////////////////////////////////


  //////////////////////////////////////////////////////////////
  // ȥåץꥹ

  hbox = gtk_hbox_new(FALSE, 8); //4ɥåȤ
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE,TRUE,0);
  gtk_widget_show(hbox);

  for (i=0; i<4; i++) {
    gtk_box_pack_start(GTK_BOX(hbox), MkStopPanel(i),TRUE,TRUE,0);
  }

  //////////////////////////////////////////////////////////////
  // ӥ͡ܥ
  hbox = gtk_hbox_new(FALSE, 4); //4ɥåȤ
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE,TRUE,0);
  gtk_widget_show(hbox);

  for (i=1;  i<=REG_NUM; i++) {
    gtk_box_pack_start(GTK_BOX(hbox), MkCombBtn(i),TRUE,TRUE,0);
  }

  // ӥ͡ܥ
  memory_button = gtk_toggle_button_new_with_label("MEM");
  gtk_box_pack_start(GTK_BOX(hbox), memory_button,TRUE,TRUE,0);
  gtk_widget_show(memory_button);

  // ȥåOFF
  button = gtk_button_new_with_label("RST");
  gtk_box_pack_start(GTK_BOX(hbox), button,TRUE,TRUE,0);
  gtk_signal_connect(GTK_OBJECT(button),"clicked",
		     GTK_SIGNAL_FUNC(unselect_all_stop),(gpointer)0);
  gtk_widget_show(button);

  //////////////////////////////////////////////////////////////////

  // chorusåܥ
  chorus_button = gtk_check_button_new_with_label("Chorus");
  gtk_box_pack_start(GTK_BOX(vbox), chorus_button, FALSE,FALSE,0);
  gtk_signal_connect(GTK_OBJECT(chorus_button),"toggled",
		     GTK_SIGNAL_FUNC(chg_chorus),(gpointer)0);
  gtk_widget_show(chorus_button);

#ifdef _WINDOWS
  gtk_widget_set_usize(main_window,560,450);
#else
  gtk_widget_set_usize(main_window,560,410);
#endif
  midi_enable_chorus(0);


#ifdef _WINDOWS
  {
    HMIDIIN  hMidiIn;
    int innum;
    int outnum;
    MIDIINCAPS InCaps;
    MIDIOUTCAPS OutCaps;
    int i;
    
    innum = midiInGetNumDevs();
    outnum = midiOutGetNumDevs();
    
    printf("\nMIDI INPUT PORTS NUM = %d\n", innum);
    for (i=0; i<innum; i++) {
      midiInGetDevCaps(i, (LPMIDIINCAPS) &InCaps,sizeof(InCaps));
      printf("  %d:%s\n",i,(LPSTR)InCaps.szPname);
    }
    printf("MIDI OUTPUT PORTS NUM = %d\n", outnum);
    for (i=0; i<outnum; i++) {
      midiOutGetDevCaps(i, (LPMIDIOUTCAPS) &OutCaps,sizeof(OutCaps));
      printf("  %d:%s\n",i,(LPSTR)OutCaps.szPname);
    }
  }
#endif

#ifdef _LINUX
  pthread_create( &thread_midi, NULL,
                  (void*(*)(void*))&midi_main_loop, (void*)NULL);
  pthread_create( &thread_js, NULL,
                  (void*(*)(void*))&joystick_main, (void*)NULL);  
#endif
  
  set_stop_name_list();
  regs_call_registration(0);
  display_combination(); 
  midi_set_preset();

  gtk_widget_show(main_window);

  gdk_threads_enter();
  gtk_main();
  gdk_threads_leave();

  return 0;
}

