/*********************************************************
  Virtual Pipe Organ for Linux (registration managiment)
  By Atsushi Ueda <a-ueda@japan-net.ne.jp>
**********************************************************/


#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "common.h"
#include "regs.h"
#include "sflist.h"
#include "midi.h"

#define VERSION "20021228"
#define VERSION_OLD1 "20020825"

static KB_STOPS div_stops[KB_NUM]; // ȥå֡٥ƥ
char soundfont_name[4096]; //sound font ̾ (ե INFOꥹINAM󥯤ꤵƤ̾Ρե̾ǤϤʤ)

// ӥ͡󡦥꡼
// 0Ͻλ¸ 1REG_NUMprogrammable
static COMBINATION_MEMORY mem[REG_NUM+1];

static char timidity_cmdline[10000]; // timidity command line


/*
  Initialization
*/
void regs_init()
{
  int i,j,k;
  for (i=0; i<KB_NUM; i++) {
    for (k=0; k<STOP_NUM; k++) {
      div_stops[i].stops[k].selected = 0;
      div_stops[i].stops[k].bank = 255;
      div_stops[i].stops[k].preset = 255;
      div_stops[i].stops[k].velocity = 63;
      div_stops[i].stops[k].stop_name[0] = 0;
    }
    div_stops[i].flg_tremulant = 0;
    strcpy(div_stops[i].werk_name, "none");
    for (k=0; k<KB_NUM; k++) {
      div_stops[i].coupler_sts[k] = 0;
    }

    for (j=0; j<=REG_NUM; j++) {
      for (k=0; k<STOP_NUM; k++) {
	mem[j].velocity[i][k] = 63;
	mem[j].selected[i][k] = 0;
      }
      mem[j].tremulant[i] = 0;
      for (k=0; k<KB_NUM; k++) {
	mem[j].coupler[i][k] = 0;
      }
    }
  }
}

char* regs_get_timidity_cmdline()
{
  return timidity_cmdline;
}

void regs_set_timidity_cmdline(char *cmdline)
{
  strcpy(timidity_cmdline, cmdline);
}


/*
  ץ֤
*/
int regs_get_current_coupler_state(int div, int to_div)
{
  if (div == to_div) {
    return 0;
  }
  return div_stops[div].coupler_sts[to_div];
}

/*
  ץ֤
*/
void regs_set_current_coupler_state(int div, int to_div, int sts)
{
  div_stops[div].coupler_sts[to_div] = sts;
}

/*
  ȥߥȤ֤
*/
int regs_get_current_tremulant_state(int div)
{
  return div_stops[div].flg_tremulant;
}

int regs_set_current_tremulant_state(int div, int sts)
{
  div_stops[div].flg_tremulant = sts;
  return 0;
}


/*
  ȥå()
*/
void regs_insert(int div, int stop)
{
  int i,j;
  for (i=STOP_NUM-1 ; i>stop; i--) {
    div_stops[div].stops[i]     = div_stops[div].stops[i-1];
    for (j=0; j<=REG_NUM; j++) {
      mem[j].selected[div][i] = mem[j].selected[div][i-1];
      mem[j].velocity[div][i] = mem[j].velocity[div][i-1];
    }
  }
  div_stops[div].stops[stop].bank     = 255;
  div_stops[div].stops[stop].preset   = 255;
  div_stops[div].stops[stop].selected = 0;
  div_stops[div].stops[stop].velocity = DEFAULT_VELOCITY;
  for (j=0; j<=REG_NUM; j++) {
    mem[j].selected[div][stop] = 0;
    mem[j].velocity[div][stop] = 0;
  }
  regs_set_stop_name_all();  
}


/*
  ȥå
*/
void regs_delete(int div, int stop)
{
  int i,j;
  for (i=stop; i<STOP_NUM-1; i++) {
    div_stops[div].stops[i]     = div_stops[div].stops[i+1];
    for (j=0; j<=REG_NUM; j++) {
      mem[j].selected[div][i] = mem[j].selected[div][i+1];
      mem[j].velocity[div][i] = mem[j].velocity[div][i+1];
    }
  }
  div_stops[div].stops[STOP_NUM-1].bank     = 255;
  div_stops[div].stops[STOP_NUM-1].preset   = 255;
  div_stops[div].stops[STOP_NUM-1].selected = 0;
  div_stops[div].stops[STOP_NUM-1].velocity = DEFAULT_VELOCITY;
  for (j=0; j<=REG_NUM; j++) {
    mem[j].selected[div][STOP_NUM-1] = 0;
    mem[j].velocity[div][STOP_NUM-1] = 0;
  }
  regs_set_stop_name_all();  
}


/*
*/
void regs_get_preset(int div, int stop, int *bank, int *preset)
{
  *bank = div_stops[div].stops[stop].bank;
  *preset = div_stops[div].stops[stop].preset;
}

void regs_set_preset(int div, int stop, int bank, int preset)
{
  div_stops[div].stops[stop].bank = bank;
  div_stops[div].stops[stop].preset = preset;
  regs_set_stop_name_all();
}



/*
  ̾
*/
char* regs_get_werk_name(int keyboard)
{
  return div_stops[keyboard].werk_name;
}

/*
  ̾
*/
void  regs_set_werk_name(int keyboard, const char *name)
{
  strncpy(div_stops[keyboard].werk_name, name, 63);
}



/*
    ȥå֤ 
    keyboard:ֹ (0~3)
    stopno:ȥåֹ(0~15) 
     0:OFF 1:ON 

    Get stop status
    keyboard: keyboard number (0~3)
    stopno: stop number (0~15) 
    return value 0:ON 1:OFF
 */
int regs_get_current_stop_select(int keyboard, int stopno)
{
  return div_stops[keyboard].stops[stopno].selected;
}

void regs_set_current_stop_select(int keyboard, int stopno, int select)
{
  div_stops[keyboard].stops[stopno].selected = select;
}


int regs_get_current_stop_velocity(int keyboard, int stopno)
{
  return div_stops[keyboard].stops[stopno].velocity;
}

void regs_set_current_stop_velocity(int keyboard, int stopno, int velocity)
{
  div_stops[keyboard].stops[stopno].velocity = velocity;
}


char* regs_get_stop_name(int keyboard, int stopno)
{
  return div_stops[keyboard].stops[stopno].stop_name;
}

void regs_set_stop_name(int keyboard, int stopno, char *name)
{
  strncpy(div_stops[keyboard].stops[stopno].stop_name, name?name:"", 60);
}

void regs_call_registration(int regno)
{
  int i,j;

  if (regno > REG_NUM) {
    return;
  }

  for (i=0; i<KB_NUM; i++) {
    for (j=0; j<STOP_NUM; j++) {
      regs_set_current_stop_select(i, j, mem[regno].selected[i][j]);
      //regs_set_current_stop_velocity(i, j, mem[regno].velocity[i][j]);
    }
    for (j=0; j<KB_NUM; j++) {
      regs_set_current_coupler_state(i,j,mem[regno].coupler[i][j]);
    }   
    regs_set_current_tremulant_state(i,mem[regno].tremulant[i]);
  }
}

void regs_memory_registration(int regno)
{
  int i,j;
  for (i=0; i<KB_NUM; i++) {
    for (j=0; j<STOP_NUM; j++) {
      mem[regno].selected[i][j] = regs_get_current_stop_select  (i, j);
      mem[regno].velocity[i][j] = regs_get_current_stop_velocity(i, j);
    }
    for (j=0; j<KB_NUM; j++) {
      mem[regno].coupler[i][j] = regs_get_current_coupler_state(i,j);
    }   
    mem[regno].tremulant[i] = regs_get_current_tremulant_state(i);
  }
}





void aaa()
{
  int i,j;
  for (i=0; i<4; i++) {
    for (j=0; j<15; j++) {
      printf("%d,%d\t", i,j);
      printf("%s\t", regs_get_stop_name(i,j));
      printf("\n");
    }
  }
}

void regs_set_stop_name_all()
{
  int i,j;
  for (i=0; i<4; i++) {
    for (j=0; j<STOP_NUM; j++) {
      regs_set_stop_name(i,j, sflist_getname(div_stops[i].stops[j].bank,
					     div_stops[i].stops[j].preset));
    }
  }
}


/*
  쥸ȥ졼եɤ߹
  fn : ե̾
 */
int regs_read_registration_table_file(char* table_fn)
{
  int i,j,k;
  char strtmp[10000];
  int flgNotDivButKb = 0;

  GetPrivateProfileString("General","Version",VERSION_OLD1,strtmp,1000,table_fn);
  if (strcmp(strtmp, VERSION_OLD1)<=0) {
    flgNotDivButKb=1;
  }

  GetPrivateProfileString("General","SoundFont","",strtmp,1000,table_fn);
  csv_read(strtmp, soundfont_name, 0); //to remove double quotation

  GetPrivateProfileString("General","ChorusLevel","750",strtmp,32,table_fn);
  midi_set_chorus_level(atoi(strtmp));

  GetPrivateProfileString("General","pitch","0.00",strtmp,32,table_fn);
  midi_set_pitch_level(atof(strtmp));

  for (i=0; i<4; i++) {
    char section[64], key[64], res[640];
    sprintf(section, flgNotDivButKb ? "kb%d" : "div%d", i);
    GetPrivateProfileString(section, "name", "none", res, 63, table_fn);
    regs_set_werk_name(i, res);

    for (j=0; j<STOP_NUM; j++) {
      char f0[32],f1[32],f2[32],f3[32];
      sprintf(key, "stop%d", j);
      GetPrivateProfileString(section, key, "255,255,0,0", res, 63, table_fn);
      csv_read(res, f0, 0);
      csv_read(res, f1, 1);
      csv_read(res, f2, 2);
      csv_read(res, f3, 3);
      regs_set_preset(i,j, atoi(f0),atoi(f1));
      regs_set_current_stop_select(i,j,atoi(f3));
      regs_set_current_stop_velocity(i,j,atoi(f2));
    }

    regs_set_stop_name_all();

    // combination memory (stops)
    for (j=0; j<=REG_NUM; j++) {
      sprintf(key,"reg%d", j);
      GetPrivateProfileString(section, key , "", res, 639, table_fn);
      for (k=0; k<STOP_NUM; k++) {
	char tmp1[32],tmp2[32];
	csv_read(res, tmp1, k*2);
	csv_read(res, tmp2, k*2+1);
	mem[j].velocity[i][k]=atoi(tmp2);
	mem[j].selected[i][k]=atoi(tmp1);
      }
    }

    // combination memory (coupler)
    for (j=0; j<=REG_NUM; j++) {
      sprintf(key,"reg_coupler%d", j);
      GetPrivateProfileString(section, key , "", res, 639, table_fn);
      for (k=0; k<KB_NUM; k++) {
	char tmp1[32];
	csv_read(res, tmp1, k);
	mem[j].coupler[i][k]=atoi(tmp1);
      }
    }

    // combination memory (tremulant)
    for (j=0; j<=REG_NUM; j++) {
      sprintf(key,"reg_tremulant%d", j);
      GetPrivateProfileString(section, key , "", res, 639, table_fn);
      mem[j].tremulant[i] = atoi(res);
    }
  }
  return TRUE;
}

/*
  쥸ȥ졼եν񤭹
  fn : ե̾
 */
int regs_write_registration_table_file(char* table_fn)
{
  int i,j,k, len;
  char str[640], section[64], key[64];

  WritePrivateProfileString("General", "Version", VERSION, table_fn);
  sprintf(str, "\x22%s\x22", soundfont_name);
  WritePrivateProfileString("General", "SoundFont", str, table_fn);
  sprintf(str, "%d", midi_get_chorus_level());
  WritePrivateProfileString("General", "ChorusLevel", str, table_fn);
  sprintf(str, "%1.2f", midi_get_pitch_level());
  WritePrivateProfileString("General", "Pitch", str, table_fn);

  for (i=0; i<4; i++) {
    sprintf(section, "div%d", i);
    WritePrivateProfileString(section, "name", 
			      regs_get_werk_name(i), 
			      table_fn);
    // current stop selection
    for (j=0; j<STOP_NUM; j++) {
      int bank,preset;
      sprintf(key, "stop%d", j);
      regs_get_preset(i,j,&bank,&preset);
      sprintf(str,"%d,%d,%d,%d",bank,preset,
	      regs_get_current_stop_velocity(i,j),
	      regs_get_current_stop_select(i,j));
      WritePrivateProfileString(section,key,str,table_fn);
    }
    // combination memory (stops)
    for (j=0; j<=REG_NUM; j++) {
      sprintf(key,"reg%d", j);
      len = 0;
      for (k=0; k<STOP_NUM; k++) {
	len += sprintf(str+len, "%d,%d",
		       mem[j].selected[i][k],
		       mem[j].velocity[i][k]);
	if (k!=STOP_NUM-1) {
	  len += sprintf(str+len, ",");
	}
      }
      WritePrivateProfileString(section,key,str,table_fn);
    }
    // combination memory (coupler)
    for (j=0; j<=REG_NUM; j++) {
      sprintf(key,"reg_coupler%d", j);
      len=0;
      for (k=0; k<KB_NUM; k++) {
	len += sprintf(str+len, "%d", mem[j].coupler[i][k]);
	if (k!=KB_NUM-1) {
	  len += sprintf(str+len, ",");
	}
      }
      WritePrivateProfileString(section,key,str,table_fn);
    }
    // combination memory (tremulant)
    for (j=0; j<=REG_NUM; j++) {
      sprintf(key, "reg_tremulant%d", j);
      sprintf(str, "%d", mem[j].tremulant[i]);
      WritePrivateProfileString(section,key,str,table_fn);
    }
  }

  return TRUE;
}


char *regs_get_soundfont_name()
{
  return soundfont_name;
}

void regs_set_soundfont_name(char *sfname)
{
  strcpy(soundfont_name, sfname);
}
